#!/usr/bin/env python
import sys
import pytest

if __name__ == '__main__':
    # show output results from every test function
    args = ['-v', 'test.py']
    # show the message output for skipped and expected failure tests
    args.append('-rxs')
    args.extend(sys.argv[1:])
    # call pytest and exit with the return code from pytest so that
    # travis will fail correctly if tests fail
    sys.exit(pytest.main(args))



import sys


try:
    from stdlib_list import stdlib_list
    pyver = '%s.%s' % (sys.version_info.major, sys.version_info.minor)
    builtin_modules = stdlib_list(pyver)
    del pyver
except ImportError:
    # assuming py>=3.10
    builtin_modules = list(set(list(sys.stdlib_module_names) + list(sys.builtin_module_names)))



# Copyright (c) <2015-2016>, Eric Dill
#
# All rights reserved.  Redistribution and use in source and binary forms, with
# or without modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import print_function, division, absolute_import

import ast
import logging
import os
import sys
from collections import defaultdict
from typing import Union

from .stdliblist import builtin_modules

from .utils import (
    AST_QUESTIONABLE,
    namespace_packages,
    SKETCHY_TYPES_TABLE,
)

logger = logging.getLogger('depfinder')


PACKAGE_NAME = None
STRICT_CHECKING = False


def get_top_level_import_name(name, custom_namespaces=None):
    num_dot = name.count(".")
    custom_namespaces = custom_namespaces or []

    if name in namespace_packages or name in custom_namespaces or name in builtin_modules:
        return name
    elif any(
        ((num_dot - nsp.count(".")) == 1) and name.startswith(nsp + ".")
        for nsp in custom_namespaces
    ):
        # this branch happens when name is foo.bar.baz and the namespace is
        # foo.bar
        return name
    else:
        if '.' not in name:
            return name
        else:
            return get_top_level_import_name(
                name.rsplit('.', 1)[0],
                custom_namespaces=custom_namespaces
            )


class ImportFinder(ast.NodeVisitor):
    """Find all imports in an Abstract Syntax Tree (AST).

    Attributes
    ----------
    required_modules : list
        The list of imports that were found outside of try/except blocks,
        function definitions and class definitions
    sketchy_modules : list
        The list of imports that were found inside of try/except blocks,
        function definitions and class definitions
    imports : list
        The list of all ast.Import nodes in the AST
    import_froms : list
        The list of all ast.ImportFrom nodes in the AST

    """

    def __init__(self, filename='', custom_namespaces=None):
        self.filename = filename
        self.required_modules = set()
        self.sketchy_modules = set()
        self.builtin_modules = set()
        self.relative_modules = set()
        self.imports = []
        self.import_froms = []
        self.total_imports = defaultdict(dict)
        self.sketchy_nodes = {}
        self.custom_namespaces = custom_namespaces or []
        super(ImportFinder, self).__init__()

    def visit(self, node):
        """Recursively visit all ast nodes.

        Look for Import and ImportFrom nodes. Classify them as being imports
        that are built in, relative, required or questionable. Questionable
        imports are those that occur within the context of a try/except block, a
        function definition or a class definition.

        Parameters
        ----------
        node : ast.AST
            The node to start the recursion
        """
        # add the node to the try/except block to signify that
        # something potentially odd is going on in this import
        if isinstance(node, AST_QUESTIONABLE):
            self.sketchy_nodes[node] = node
        super(ImportFinder, self).visit(node)
        # after the node has been recursed in to, remove the try node
        self.sketchy_nodes.pop(node, None)

    def visit_Import(self, node: ast.Import):
        """Executes when an ast.Import node is encountered

        an ast.Import node is something like 'import bar'

        If ImportCatcher is inside of a try block then the import that has just
        been encountered will be added to the `sketchy_modules` instance
        attribute. Otherwise the module will be added to the `required_modules`
        instance attribute
        """
        self.imports.append(node)
        self._add_to_total_imports(node)

        mods = set([
            get_top_level_import_name(name.name, custom_namespaces=self.custom_namespaces)
            for name in node.names
        ])
        for mod in mods:
            self._add_import_node(mod)

    def visit_ImportFrom(self, node: ast.ImportFrom):
        """Executes when an ast.ImportFrom node is encountered

        an ast.ImportFrom node is something like 'from foo import bar'

        If ImportCatcher is inside of a try block then the import that has just
        been encountered will be added to the `sketchy_modules` instance
        attribute. Otherwise the module will be added to the `required_modules`
        instance attribute
        """
        self.import_froms.append(node)
        if node.module is None:
            # this is a relative import like 'from . import bar'
            # so do nothing
            return
        if node.level > 0:
            # this is a relative import like 'from .foo import bar'
            node_name = get_top_level_import_name(
                node.module, custom_namespaces=self.custom_namespaces
            )
            self.relative_modules.add(node_name)
            return
        # this is a non-relative import like 'from foo import bar'
        self._add_to_total_imports(node)
        node_name = get_top_level_import_name(
            node.module,
            custom_namespaces=self.custom_namespaces,
        )
        self._add_import_node(node_name)

    def _add_to_total_imports(self, node: Union[ast.Import, ast.ImportFrom]):
        import_metadata = {}
        try:
            import_metadata.update({'exact_line': ast.unparse(node)})
        except AttributeError:
            pass

        import_metadata.update({v: False for v in SKETCHY_TYPES_TABLE.values()})
        import_metadata.update({SKETCHY_TYPES_TABLE[node.__class__]: True for node in self.sketchy_nodes})
        names = set()
        if isinstance(node, ast.Import):
            _names = set(name.name for name in node.names)
            import_metadata['import'] = _names
            names.update(_names)
        elif isinstance(node, ast.ImportFrom):
            import_metadata['import_from'] = {node.module}
            names.add(node.module)
        else:
            raise NotImplementedError(f"Expected ast.Import or ast.ImportFrom this is {type(node)}")

        for name in names:
            self.total_imports[name].update({(self.filename, node.lineno): import_metadata})

    def _add_import_node(self, node_name):
        # see if the module is a builtin
        if node_name in builtin_modules:
            self.builtin_modules.add(node_name)
            return

        # see if we are in a try block
        if self.sketchy_nodes:
            self.sketchy_modules.add(node_name)
            return

        # if none of the above cases are true, it is likely that this
        # ImportFrom node occurs at the top level of the module
        self.required_modules.add(node_name)

    def describe(self):
        """Return the found imports

        Returns
        -------
        dict :
            'required': The modules that were encountered outside of a
                        try/except block
            'questionable': The modules that were encountered inside of a
                            try/except block
            'relative': The modules that were imported via relative import
                        syntax
            'builtin' : The modules that are part of the standard library
        """
        desc = {
            'required': self.required_modules,
            'relative': self.relative_modules,
            'questionable': self.sketchy_modules,
            'builtin': self.builtin_modules
        }
        desc = {k: v for k, v in desc.items() if v}
        return desc

    def __repr__(self):
        return 'ImportCatcher: %s' % repr(self.describe())


def get_imported_libs(code, filename='', custom_namespaces=None):
    """Given a code snippet, return a list of the imported libraries

    Parameters
    ----------
    code : str
        The code to parse and look for imports

    Returns
    -------
    ImportCatcher
        The ImportCatcher is the object in `depfinder` that contains all the
        information regarding which imports were found where.  You will most
        likely be interested in calling the describe() function on this return
        value.

    Examples
    --------
    >>> depfinder.get_imported_libs('from foo import bar')
    {'required': {'foo'}, 'questionable': set()}
    >>> with open('depfinder.py') as f:
            code = f.read()
            imports = depfinder.get_imported_libs(code)
            print(imports.describe())
    {'builtin': {'__future__', 'json', 'ast', 'os', 'sys', 'collections'},
     'required': {'stdlib_list'}}
    """
    # skip ipython notebook lines
    code = '\n'.join([line for line in code.split('\n')
                      if not line.startswith('%')])
    tree = ast.parse(code)
    import_finder = ImportFinder(filename=filename, custom_namespaces=custom_namespaces)
    import_finder.visit(tree)
    return import_finder


def parse_file(python_file, custom_namespaces=None):
    """Parse a single python file

    Parameters
    ----------
    python_file : str
        Path to the python file to parse for imports

    Returns
    -------
    catchers : tuple
        Yields tuples of (module_name, full_path_to_module, ImportCatcher)
    """
    global PACKAGE_NAME
    if PACKAGE_NAME is None:
        PACKAGE_NAME = os.path.basename(python_file).split('.')[0]
        logger.debug("Setting PACKAGE_NAME global variable to {}"
                     "".format(PACKAGE_NAME))
    # Try except block added for adal package which has a BOM at the beginning,
    # requiring a different encoding to load properly
    try:
        with open(python_file, 'r') as f:
            code = f.read()
        catcher = get_imported_libs(
            code, filename=python_file, custom_namespaces=custom_namespaces
        )
    except SyntaxError:
        with open(python_file, 'r', encoding='utf-8-sig') as f:
            code = f.read()
        catcher = get_imported_libs(
            code, filename=python_file, custom_namespaces=custom_namespaces
        )
    catcher.total_imports = dict(catcher.total_imports)
    mod_name = os.path.split(python_file)[:-3]
    return mod_name, python_file, catcher


def iterate_over_library(path_to_source_code, custom_namespaces=None):
    """Helper function to recurse into a library and find imports in .py files.

    This allows the user to apply filters on the user-side to exclude imports
    based on their file names.
    `conda-skeletor <https://github.com/ericdill/conda-skeletor>`_
    makes heavy use of this function

    Parameters
    ----------
    path_to_source_code : str

    Yields
    -------
    catchers : tuple
        Yields tuples of (module_name, full_path_to_module, ImportCatcher)
    """
    global PACKAGE_NAME
    global STRICT_CHECKING
    if PACKAGE_NAME is None:
        PACKAGE_NAME = os.path.basename(path_to_source_code).split('.')[0]
        logger.debug("Setting PACKAGE_NAME global variable to {}"
                     "".format(PACKAGE_NAME))
    skipped_files = []
    all_files = []
    for parent, folders, files in os.walk(path_to_source_code):
        for f in files:
            if f.endswith('.py'):
                full_file_path = os.path.join(parent, f)
                all_files.append(full_file_path)
                try:
                    yield parse_file(full_file_path, custom_namespaces=custom_namespaces)
                except Exception:
                    logger.exception("Could not parse file: {}".format(full_file_path))
                    skipped_files.append(full_file_path)
    if skipped_files:
        logger.warning("Skipped {}/{} files".format(len(skipped_files), len(all_files)))
        for idx, f in enumerate(skipped_files):
            logger.warn("%s: %s" % (str(idx), f))

    if skipped_files and STRICT_CHECKING:
        raise RuntimeError("Some files failed to parse. See logs for full stack traces.")



# Copyright (c) <2015-2016>, Eric Dill
#
# All rights reserved.  Redistribution and use in source and binary forms, with
# or without modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import print_function, division, absolute_import

import copy
import io
import json
import logging
from collections import defaultdict
from fnmatch import fnmatch

from .inspection import iterate_over_library, get_imported_libs
from .utils import pkg_data

logger = logging.getLogger('depfinder')

STRICT_CHECKING = False


def simple_import_search(path_to_source_code, remap=True, ignore=None, custom_namespaces=None):
    """Return all imported modules in all .py files in `path_to_source_code`

    Parameters
    ----------
    path_to_source_code : str
    remap : bool, optional
        Normalize the import names to be synonymous with their conda/pip names
    ignore : list, optional
        String pattern which if matched causes the file to not be inspected
    custom_namespaces : list of str or None
        If not None, then resulting package outputs will list everying under these
        namespaces (e.g., for packages foo.bar and foo.baz, the outputs are foo.bar
        and foo.baz instead of foo if custom_namespaces=["foo"]).

    Returns
    -------
    dict
        The list of all imported modules, sorted according to the keys listed
        in the docstring of depfinder.ImportCatcher.describe()

    Examples
    --------
    >>> depfinder.simple_import_search('/path/to/depfinder/source')
    {'builtin': ['__future__',
                 'ast',
                 'collections',
                 'json',
                 'os',
                 'shlex',
                 'sys',
                 'tempfile'],
     'required': ['depfinder',
                  'nbformat',
                  'pytest',
                  'setuptools',
                  'sphinx_rtd_theme',
                  'stdlib_list',
                  'test_with_code']}
    """
    all_deps = defaultdict(set)
    catchers = iterate_over_library(path_to_source_code, custom_namespaces=custom_namespaces)
    for mod, path, catcher in catchers:
        # if ignore provided skip things which match the ignore pattern
        if ignore and any(fnmatch(path, i) for i in ignore):
            continue
        for k, v in catcher.describe().items():
            all_deps[k].update(v)

    all_deps = {k: sorted(list(v)) for k, v in all_deps.items() if v}
    if remap:
        return sanitize_deps(all_deps)
    return all_deps


def notebook_path_to_dependencies(path_to_notebook, remap=True, custom_namespaces=None):
    """Helper function that turns a jupyter notebook into a list of dependencies

    Parameters
    ----------
    path_to_notebook : str
    remap : bool, optional
        Normalize the import names to be synonymous with their conda/pip names
    custom_namespaces : list of str ot None, optional
        If not None, then resulting package outputs will list everying under these
        namespaces (e.g., for packages foo.bar and foo.baz, the outputs are foo.bar
        and foo.baz instead of foo if custom_namespaces=["foo"]).

    Returns
    -------
    dict
        Dict of dependencies keyed on

        - 'builtin' - libraries built in to python
        - 'required' - libraries that are found at the top level of your modules
        - 'questionable' - libraries that are found inside try/except blocks
        - 'relative' - libraries that are relative imports

    Examples
    --------
    >>> depfinder.notebook_path_to_dependencies('depfinder_usage.ipynb')
    {'builtin': ['os', 'pprint'], 'required': ['depfinder']}
    """
    try:
        from IPython.core.inputsplitter import IPythonInputSplitter
        transform = IPythonInputSplitter(line_input_checker=False).transform_cell
    except:
        transform = lambda code: code

    nb = json.load(io.open(path_to_notebook, encoding='utf8'))
    codeblocks = [''.join(cell['source']) for cell in nb['cells']
                  if cell['cell_type'] == 'code']
    all_deps = defaultdict(set)

    for codeblock in codeblocks:
        codeblock = transform(codeblock)
        # TODO this may fail on py2/py3 syntax when running in the other runtime.
        # May want to consider updating some error handling around that case.
        # Will wait until that use case surfaces before modifying
        deps_dict = get_imported_libs(codeblock, custom_namespaces=custom_namespaces).describe()
        for k, v in deps_dict.items():
            all_deps[k].update(v)

    all_deps = {k: sorted(list(v)) for k, v in all_deps.items()}
    if remap:
        return sanitize_deps(all_deps)
    return all_deps


def sanitize_deps(deps_dict):
    """
    Helper function that takes the output of `notebook_path_to_dependencies`
    or `simple_import_search` and turns normalizes the import names to be
    synonymous with their conda/pip names

    Parameters
    ----------
    deps_dict : dict
        Output of `notebook_path_to_dependencies` or `simple_import_search`
    Returns
    -------
    deps_dict : dict
        If remap is True: Sanitized `deps_dict`
        If remap is False: `deps_dict`
    """
    from .inspection import PACKAGE_NAME
    new_deps_dict = {}
    list_of_possible_fakes = set([v for val in pkg_data['_FAKE_PACKAGES'].values() for v in val])
    for k, packages_list in deps_dict.items():

        pkgs = copy.copy(packages_list)
        new_deps_dict[k] = set()
        for pkg in pkgs:
            # drop fake packages
            if pkg in list_of_possible_fakes:
                logger.debug("Ignoring {} from the list of imports. It is "
                             "installed as part of another package. Set the "
                             "`--no-remap` cli flag if you want to disable "
                             "this".format(pkg))
                continue
            if pkg == PACKAGE_NAME:
                logger.debug("Ignoring {} from the list of imports. It is "
                             "the name of the package that we are trying to "
                             "find the dependencies for. Set the `--no-remap` "
                             "cli flag if you want to disable this.".format(pkg))
                continue
            pkg_to_add = pkg_data['_PACKAGE_MAPPING'].get(pkg, pkg)
            if pkg != pkg_to_add:
                logger.debug("Renaming {} to {}".format(pkg, pkg_to_add))
            new_deps_dict[k].add(pkg_to_add)
    new_deps_dict = {k: sorted(list(v)) for k, v in new_deps_dict.items() if v}
    return new_deps_dict


def simple_import_search_conda_forge_import_map(path_to_source_code, builtins=None, ignore=None, custom_namespaces=None):
    """Return all conda-forge packages used in all .py files in `path_to_source_code`

    Parameters
    ----------
    path_to_source_code : str
    builtins : list, optional
        List of python builtins to partition into their own section
    ignore : list, optional
        String pattern which if matched causes the file to not be inspected
    custom_namespaces : list of str or None
        If not None, then resulting package outputs will list everying under these
        namespaces (e.g., for packages foo.bar and foo.baz, the outputs are foo.bar
        and foo.baz instead of foo if custom_namespaces=["foo"]).

    Returns
    -------
    dict
        The list of all imported modules, sorted according to the keys listed
        in the docstring of depfinder.ImportCatcher.describe()

    Examples
    --------
    >>> depfinder.simple_import_search_conda_forge_import_map('/path/to/depfinder/source')
    {'builtin': ['__future__',
                 'ast',
                 'collections',
                 'json',
                 'os',
                 'shlex',
                 'sys',
                 'tempfile'],
     'required': ['depfinder',
                  'nbformat',
                  'pytest',
                  'setuptools',
                  'sphinx_rtd_theme',
                  'stdlib_list',
                  'test_with_code']}
    """
    # run depfinder on source code
    if ignore is None:
        ignore = []
    total_imports_list = []
    for _, _, c in iterate_over_library(path_to_source_code, custom_namespaces=custom_namespaces):
        total_imports_list.append(c.total_imports)
    total_imports = defaultdict(dict)
    for total_import in total_imports_list:
        for name, md in total_import.items():
            total_imports[name].update(md)
    from .reports import report_conda_forge_names_from_import_map
    imports, _, _ = report_conda_forge_names_from_import_map(
        total_imports, builtin_modules=builtins, ignore=ignore
    )
    return {k: sorted(list(v)) for k, v in imports.items()}


def simple_import_to_pkg_map(path_to_source_code, builtins=None, ignore=None, custom_namespaces=None):
    """Provide the map beteen all the imports and their possible packages

    Parameters
    ----------
    path_to_source_code : str
    builtins : list, optional
        List of python builtins to partition into their own section
    ignore : list, optional
        String pattern which if matched causes the file to not be inspected
    custom_namespaces : list of str or None
        If not None, then resulting package outputs will list everying under these
        namespaces (e.g., for packages foo.bar and foo.baz, the outputs are foo.bar
        and foo.baz instead of foo if custom_namespaces=["foo"]).

    Returns
    -------
    dict of dict of sets:
        The classes of requirements (required, questionable, builtin, no match required, no match questionable),
        name of the import and packages that provide that import

    """
    # run depfinder on source code
    if ignore is None:
        ignore = []
    total_imports_list = []
    for _, _, c in iterate_over_library(path_to_source_code, custom_namespaces=custom_namespaces):
        total_imports_list.append(c.total_imports)
    total_imports = defaultdict(dict)
    for total_import in total_imports_list:
        for name, md in total_import.items():
            total_imports[name].update(md)
    from .reports import report_conda_forge_names_from_import_map
    _, _, import_to_artifact = report_conda_forge_names_from_import_map(
        total_imports, builtin_modules=builtins, ignore=ignore
    )
    return import_to_artifact



from __future__ import (unicode_literals, print_function, division,
                        absolute_import)

import contextlib
import itertools
import os
import random
import subprocess
import sys
import tempfile
from collections import defaultdict
from os.path import dirname, join

import pytest
import six
from nbformat import v4

import depfinder
from depfinder import cli, main, inspection, parse_file
from depfinder.main import simple_import_search_conda_forge_import_map, simple_import_to_pkg_map
from depfinder.reports import report_conda_forge_names_from_import_map, extract_pkg_from_import, \
    recursively_search_for_name, _builtin_modules

random.seed(12345)

# Testing spec:
# - targets: dict
#   - required: Iterable of library imports that should be found
#   - questionable: Iterable of library imports that should be found
# - code: String. Code that will be parsed to look for imports


complex_imports = [
    {'targets':
         {'questionable': ['atom', 'chemist', 'molecule', 'physicist']},
     'code': """
try:
    import molecule
except ImportError:
    import atom
else:
    import chemist
finally:
    import physicist"""
     },
    {'targets': {'required': ['foo'], 'builtin': ['os']},
     'code': """
import foo
try:
    import os
except ImportError:
    # why would you put this in a try block??
    pass"""},
    {'targets': {'required': ['toplevel', 'toplevelfrom'],
                 'questionable': ['function_inside_class',
                                  'function_inside_class_from',
                                  'inside_class',
                                  'inside_class_from',
                                  'inside_function',
                                  'inside_function_from'],
                 'relative': ['relative_function',
                              'relative_function_inside_class',
                              'relative_inside_class'],
                 'builtin': ['os', 'pprint']},
     'code': """
from toplevelfrom import some_function
import toplevel
def function():
    import inside_function
    from inside_function_from import another_function
    from .relative_function import random_function
class Class:
    import inside_class
    from inside_class_from import some_function
    from .relative_inside_class import a_third_function
    import os
    def __init__(self):
        import function_inside_class
        from function_inside_class_from import some_function
        from .relative_function_inside_class import a_third_function
        import pprint
"""},

    {'targets':
         {'questionable': ['chico', 'groucho', 'harpo']},
     'code': """
if this:
    import groucho
elif that:
    import harpo
else:
    import chico"""
     },
]

simple_imports = [
    {'targets': {'required': ['foo']},
     'code': 'import foo'},
    {'targets': {'required': ['bar', 'foo']},
     'code': 'import foo, bar'},
    {'targets': {'required': ['numpy']},
     'code': 'import numpy'},
    {'targets': {'required': ['matplotlib']},
     'code': 'from matplotlib import pyplot'},
    # Hit the fake packages code block in main.sanitize_deps()
    {'targets': {'required': ['numpy']},
     'code': 'from numpy import warnings as npwarn'},
]

relative_imports = [
    {'targets': {},
     'code': 'from . import bar'},
    {'targets': {'relative': ['bar']},
     'code': 'from .bar import baz'},
    {'targets': {'relative': ['bar']},
     'code': 'from ..bar import baz'},
]

@pytest.fixture(scope='module')
def using_stdlib_list():
    try:
        import stdlib_list
        return True
    except ImportError:
        return False

def test_nested_namespace_builtins(using_stdlib_list):
    if using_stdlib_list:
        expected = {'builtin': ['concurrent.futures']}
    else:
        expected = {'builtin': ['concurrent']}
    code = 'import concurrent.futures'



    test_object = Initter({'targets': expected, 'code': code})
    imports = main.get_imported_libs(test_object.code)
    assert imports.describe() == test_object.targets


class Initter(object):
    def __init__(self, artifact):
        targets = artifact.get('targets', {})
        self.targets = {k: set(v) for k, v in targets.items()}
        self.code = artifact['code']


def test_imports():
    for simple_import in complex_imports + simple_imports:
        test_object = Initter(simple_import)
        imports = main.get_imported_libs(test_object.code)
        assert imports.describe() == test_object.targets


def test_relative_imports():
    for rel in relative_imports:
        test_object = Initter(rel)
        imports = main.get_imported_libs(test_object.code)
        assert imports.describe() == test_object.targets


def test_for_smoke():
    """Do not validate the output of the functions, just make sure that calling
    them does not make depfinder blow up
    """
    deps = list(main.iterate_over_library('.'))
    assert deps is not None
    assert str(deps) is not None
    assert repr(deps) is not None
    # hit the simple api
    assert main.simple_import_search('.') is not None


### NOTEBOOK TESTING CODE ###

@contextlib.contextmanager
def write_notebook(cells):
    nb = v4.new_notebook()
    nb['cells'] = [v4.new_code_cell(code_cell) for code_cell in cells]
    fname = tempfile.NamedTemporaryFile(suffix='.ipynb').name
    with open(fname, 'w') as f:
        f.write(v4.writes(nb))

    try:
        yield fname
    finally:
        os.remove(fname)


def test_notebook_remapping():
    code = "import mpl_toolkits"
    with write_notebook([code]) as fname:
        deps = main.notebook_path_to_dependencies(fname, remap=False)
        assert {'required': ['mpl_toolkits']} == deps
        assert {} == main.notebook_path_to_dependencies(fname)


@pytest.mark.parametrize("import_list_dict", [complex_imports,
                                              simple_imports,
                                              relative_imports])
def tester(import_list_dict, capsys):
    # http://nbviewer.ipython.org/gist/fperez/9716279
    for import_dict in import_list_dict:
        cell_code = [import_dict['code']]
        target = import_dict['targets']
        with write_notebook(cell_code) as fname:
            # parse the notebook!
            assert target == main.notebook_path_to_dependencies(fname)
            # check the notebook cli
            _run_cli(path_to_check=fname)
            stdout, stderr = capsys.readouterr()
            assert target == eval(stdout)


def test_multiple_code_cells(capsys):
    targets = defaultdict(set)
    import_list_dict = complex_imports + relative_imports + simple_imports
    # http://nbviewer.ipython.org/gist/fperez/9716279
    code_for_cells = []
    for import_dict in import_list_dict:
        code_for_cells.append(import_dict['code'])
        target = import_dict['targets']
        for k, v in target.items():
            targets[k].update(set(v))

    # turn targets into a dict of sorted lists
    targets = {k: sorted(list(v)) for k, v in targets.items()}
    with write_notebook(code_for_cells) as fname:
        # parse the notebook!
        assert targets == main.notebook_path_to_dependencies(fname)
        # check the notebook cli
        _run_cli(path_to_check=fname)
        stdout, stderr = capsys.readouterr()
        assert targets == eval(stdout)


### CLI TESTING CODE ###

def _process_args(path_to_check, extra_flags):
    """
    Parameters
    ----------
    path_to_check : str, optional
        Defaults to the directory of the depfinder package
    extra_flags : list, optional
        List of extra command line flags to pass.
        Defaults to passing nothing extra.
    """
    if path_to_check is None:
        path_to_check = dirname(depfinder.__file__)
    if isinstance(extra_flags, six.string_types):
        extra_flags = [extra_flags]
    if extra_flags is None:
        extra_flags = []

    return path_to_check, extra_flags


def _subprocess_cli(path_to_check=None, extra_flags=None):
    path_to_check, extra_flags = _process_args(path_to_check, extra_flags)
    p = subprocess.Popen(
        ['depfinder', path_to_check] + extra_flags,
        env=dict(os.environ),
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE
    )

    stdout, stderr = p.communicate()
    returncode = p.returncode
    return stdout, stderr, returncode


def _run_cli(path_to_check=None, extra_flags=None):
    """
    Helper function to run depfinder in its cli mode
    """
    path_to_check, extra_flags = _process_args(path_to_check, extra_flags)
    sys.argv = ['depfinder', path_to_check] + extra_flags
    cli.cli()
    return None


def known_flags():
    # option_strings are the things like ['-h', '--help']
    # or are empty lists if the action is a positional
    flags = [o.option_strings for o in cli._init_parser()._actions]
    # drop the empty flags (these are positional arguments only
    flags = [flag for flag in flags if flag]
    # now flatten the nested list
    flags = [flag for flag_twins in flags for flag in flag_twins]
    flags.remove('-k')
    flags.remove('--key')
    flags.remove('--pdb')
    flags.remove('--ignore')
    flags.remove('--custom-namespaces')
    flags.extend(['-k all', '-k required', '-k optional', '-k builtin',
                  '-k relative'])
    return flags


@pytest.fixture(scope="module")
def flags():
    yield known_flags()


@pytest.mark.parametrize(
    'flags',
    itertools.chain(
        [random.sample(known_flags(), i) for i in range(1, len(known_flags()))]
    )
)
def test_cli_with_random_flags(flags):
    """
    More of a fuzz test for depfinder than a unit test.

    Parameters
    ----------
    flags : list
        Random combination of valid command line flags
    """
    out, err, returncode = _subprocess_cli(extra_flags=flags)
    if returncode != 0:
        # The only thing that I know of that will exit with a nonzero status
        # is if you try to combine quiet mode and verbose mode. This handles
        # that case
        quiet = {'-q', '--quiet'}
        verbose = {'-v', '--verbose'}
        flags = set(flags)
        assert flags & quiet != set() and flags & verbose != set()
        return

    assert returncode == 0


@pytest.mark.parametrize(
    'path, req',
    ((dirname(depfinder.__file__), None),
     (join(dirname(depfinder.__file__), 'main.py'), set()))
)
def test_cli(path, req, capsys):
    """
    Test to ensure that the depfinder cli is finding the dependencies in the
    source the depfinder package that are listed in the requirements.txt file
    """
    main.PACKAGE_NAME = None
    old_argv = sys.argv
    sys.argv = ['depfinder']
    _run_cli(path_to_check=path)
    sys.argv = old_argv
    # read stdout and stderr with pytest's built-in capturing mechanism
    stdout, stderr = capsys.readouterr()
    print('stdout\n{}'.format(stdout))
    print('stderr\n{}'.format(stderr))
    if req is None:
        dependencies_file = join(dirname(dirname(depfinder.__file__)),
                                 'requirements.txt')
        dependencies = set([dep for dep in open(dependencies_file, 'r').read().split('\n') if not dep.startswith("stdlib")])
    else:
        dependencies = req
    assert dependencies == set(eval(stdout).get('required', set()))


def test_known_fail_cli(tmpdir):
    tmpfile = os.path.join(str(tmpdir), 'bad_file.txt')
    import this
    with open(tmpfile, 'w') as f:
        f.write("".join([this.d.get(this.c, this.c) for this.c in this.s]))

    with pytest.raises(RuntimeError):
        _run_cli(path_to_check=tmpfile)


def test_known_fail_cli2():
    with pytest.raises(cli.InvalidSelection):
        _run_cli(extra_flags=['-q', '-v'])


@pytest.mark.parametrize(
    'path',
    (dirname(depfinder.__file__),
     join(dirname(depfinder.__file__), 'main.py'))
)
def test_individual_args(path, flags):
    for flag in flags:
        if flag in ['-h', '--help']:
            # skip the help messages since they cause the system to exit
            continue
        _run_cli(path_to_check=path, extra_flags=[flag])
        _run_cli(path_to_check=path, extra_flags=[flag])


def test_fake_packages():
    fake_import = "import mpl_toolkits"
    imports = main.get_imported_libs(fake_import)
    assert imports.describe() == {'required': {'mpl_toolkits'}}
    assert main.sanitize_deps(imports.describe()) == {}


def test_get_top_level_import():
    name = 'this.that.something'
    top_level_name = inspection.get_top_level_import_name(name)
    assert top_level_name == 'this'

    name = 'google.cloud.storage.something'
    top_level_name = inspection.get_top_level_import_name(name)
    assert top_level_name == 'google.cloud.storage'


def test_report_conda_forge_names_from_import_map():
    m, f, c = parse_file(join(dirname(depfinder.__file__), 'utils.py'))
    report, import_to_artifact, import_to_pkg = report_conda_forge_names_from_import_map(c.total_imports)
    assert report['required'] == {'pyyaml', 'requests'}


def test_report_conda_forge_names_from_import_map_ignore():
    m, f, c = parse_file(join(dirname(depfinder.__file__), 'inspection.py'))
    report, import_to_artifact, import_to_pkg = report_conda_forge_names_from_import_map(c.total_imports,
                                                                                         ignore=['*insp*'])
    assert report['required'] == set()


def test_simple_import_search_conda_forge_import_map():
    path_to_source = dirname(depfinder.__file__)
    expected_result = sorted(list({"pyyaml", "requests"}))
    report = simple_import_search_conda_forge_import_map(path_to_source)
    assert report['required'] == expected_result


@pytest.mark.parametrize('import_name, expected_result', [
    ('six.moves', 'six'),
    ('win32com.shell', 'pywin32'),
    ('win32com', 'pywin32'),
    # this comes from cython but doesn't seem to be a real pkg
    ('refnanny.hi', 'refnanny.hi')
])
def test_extract_pkg_from_import_for_complex_imports(import_name, expected_result):
    result, _, _ = extract_pkg_from_import(import_name)
    assert result == expected_result


@pytest.mark.parametrize('import_name, expected_result', [
    ('six.moves', False),
])
def test_search_for_name(import_name, expected_result):
    builtin_name_maybe = recursively_search_for_name(import_name, _builtin_modules)
    assert builtin_name_maybe == expected_result


def test_simple_import_to_pkg_map():
    path_to_source = dirname(depfinder.__file__)
    import_to_artifact = simple_import_to_pkg_map(path_to_source)
    expected_result = {'builtin': {},
                                  'questionable': {'stdlib_list': {'stdlib-list'}, 'IPython.core.inputsplitter': {'ipython', 'autovizwidget'}},
                                  'questionable no match': {},
                                  'required': {'requests': {'apache-libcloud',
                                                            'arm_pyart',
                                                            'autovizwidget',
                                                            'dbxfs',
                                                            'google-api-core',
                                                            'google-cloud-bigquery-storage-core',
                                                            'requests'},
                                               'yaml': {'google-cloud-bigquery-storage-core', 'pyyaml'}},
                                  'required no match': {}}
    assert import_to_artifact == expected_result